import uuid

from typing import List

from fastapi import APIRouter, Security
from sqlalchemy import delete, select
from sqlalchemy.orm import Session

from ..core.security import get_current_active_user as get_user
from ..db.session import SessionFuture
from ..models.account_type import AccountType
from ..models.db_setting import DbSetting
from ..models.setting_type import SettingType
from ..models.voucher_type import VoucherType
from ..schemas.settings import (
    AccountTypesSelected,
    LockDate,
    LockInformation,
    VoucherTypesSelected,
)
from ..schemas.user import UserToken


router = APIRouter()


@router.post("", response_model=List[LockInformation])
def post(
    data_in: LockInformation,
    user: UserToken = Security(get_user, scopes=["lock-date"]),
) -> List[LockInformation]:
    id_ = uuid.uuid4()
    data_in.id_ = id_
    data_dict = data_in.dict()
    data_dict["voucher_types"] = [item["id_"] for item in data_dict["voucher_types"]]
    data_dict["account_types"] = [item["id_"] for item in data_dict["account_types"]]

    with SessionFuture() as db:
        lock_date = DbSetting(
            id_=id_,
            name=f"Lock Info {str(id_)}",
            setting_type=SettingType.VOUCHER_LOCK,
            data=data_in.dict(),
            valid_from=data_in.valid_from,
            valid_till=data_in.valid_till,
        )
        db.add(lock_date)
        db.commit()
    return get_info(db)


@router.delete("/{id_}", response_model=List[LockInformation])
def delete_route(
    id_: uuid.UUID,
    user: UserToken = Security(get_user, scopes=["lock-date"]),
) -> List[LockInformation]:
    with SessionFuture() as db:
        db.execute(delete(DbSetting).where(DbSetting.id == id_))
        db.commit()
    return get_info(db)


@router.get("", response_model=List[LockInformation])
def get(
    user: UserToken = Security(get_user),
) -> List[LockInformation]:
    with SessionFuture() as db:
        return get_info(db)


def get_info(db: Session) -> List[LockInformation]:
    data = db.execute(select(DbSetting.data).where(DbSetting.setting_type == SettingType.VOUCHER_LOCK)).scalars().all()
    return [
        LockInformation(
            id=item["id_"],
            validFrom=item["valid_from"],
            validTill=item["valid_till"],
            voucherTypes=[
                VoucherTypesSelected(
                    id=vt["id_"],
                    name=VoucherType(vt["id_"]).name.replace("_", " ").title(),
                )
                for vt in item["voucher_types"]
            ],
            accountTypes=[
                AccountTypesSelected(
                    id=at["id_"],
                    name=db.execute(select(AccountType.name).where(AccountType.id == at["id_"])).scalar_one(),
                )
                for at in item["account_types"]
            ],
            start=LockDate(days=item["start"]["days"], date=item["start"]["date_"]),
            finish=LockDate(days=item["finish"]["days"], date=item["finish"]["date_"]),
            index=item["index"],
        )
        for item in sorted(data, key=lambda d: d["index"], reverse=True)  # type: ignore[no-any-return]
    ]
