import uuid

from datetime import date, datetime
from typing import Dict, List, Tuple

import brewman.schemas.raw_material_cost as schemas

from fastapi import APIRouter, Request, Security
from sqlalchemy.orm import Session
from sqlalchemy.sql.expression import case, func, select

from ...core.security import get_current_active_user as get_user
from ...core.session import get_finish_date, get_start_date, set_period
from ...db.session import SessionFuture
from ...models.account_base import AccountBase
from ...models.batch import Batch
from ...models.cost_centre import CostCentre
from ...models.inventory import Inventory
from ...models.journal import Journal
from ...models.product import Product
from ...models.product_group import ProductGroup
from ...models.stock_keeping_unit import StockKeepingUnit
from ...models.voucher import Voucher
from ...models.voucher_type import VoucherType
from ...schemas.user import UserToken


router = APIRouter()


@router.get("", response_model=schemas.RawMaterialCost)
def report_blank(
    request: Request,
    user: UserToken = Security(get_user, scopes=["raw-material-cost"]),
) -> schemas.RawMaterialCost:
    return schemas.RawMaterialCost(
        startDate=get_start_date(request.session),
        finishDate=get_finish_date(request.session),
        body=[],
        footer=None,
    )


@router.get("/data", response_model=schemas.RawMaterialCost)
def report_data(
    request: Request,
    s: str,
    f: str,
    user: UserToken = Security(get_user, scopes=["raw-material-cost"]),
) -> schemas.RawMaterialCost:
    with SessionFuture() as db:
        body, footer = build_report(
            datetime.strptime(s, "%d-%b-%Y").date(), datetime.strptime(f, "%d-%b-%Y").date(), db
        )
        set_period(s, f, request.session)
        return schemas.RawMaterialCost(
            startDate=s,
            finishDate=f,
            body=body,
            footer=footer,
        )


@router.get("/{id_}", response_model=schemas.RawMaterialCost)
def report_id(
    id_: uuid.UUID,
    request: Request,
    s: str,
    f: str,
    user: UserToken = Security(get_user, scopes=["raw-material-cost"]),
) -> schemas.RawMaterialCost:
    with SessionFuture() as db:
        body = build_report_id(
            id_, datetime.strptime(s, "%d-%b-%Y").date(), datetime.strptime(f, "%d-%b-%Y").date(), db
        )
        set_period(s, f, request.session)
        return schemas.RawMaterialCost(
            id=id_,
            startDate=s,
            finishDate=f,
            body=body,
        )


def build_report(
    start_date: date, finish_date: date, db: Session
) -> Tuple[List[schemas.RawMaterialCostItem], schemas.RawMaterialCostItem]:
    body = []
    sum_issue = func.sum(case([(AccountBase.type_id == 2, Journal.signed_amount)], else_=0)).label("issue")
    sum_sale = func.sum(case([(AccountBase.type_id == 3, Journal.signed_amount * -1)], else_=0)).label("sale")

    query = db.execute(
        select(CostCentre, sum_issue, sum_sale)
        .join(CostCentre.journals)
        .join(Journal.voucher)
        .join(Journal.account)
        .where(
            Voucher.date >= start_date,
            Voucher.date <= finish_date,
            Journal.cost_centre_id != CostCentre.cost_centre_purchase(),
            AccountBase.type_id.in_([2, 3]),
        )
        .group_by(CostCentre)
        .order_by(sum_sale.desc())
    ).all()

    issues = 0
    sales = 0
    for cost_centre, issue, sale in query:
        issues += issue
        sales += sale
        rmc = 0 if sale == 0 else issue / sale
        body.append(
            schemas.RawMaterialCostItem(
                name=cost_centre.name,
                issue=issue,
                sale=sale,
                rmc=rmc,
                url=["/", "raw-material-cost", str(cost_centre.id)],
                order=0,
            )
        )

    rmc = 0 if sales == 0 else issues / sales
    return body, schemas.RawMaterialCostItem(name="Total", issue=issues, sale=sales, rmc=rmc, order=0)


def build_report_id(
    cost_centre_id: uuid.UUID, start_date: date, finish_date: date, db: Session
) -> List[schemas.RawMaterialCostItem]:
    sum_quantity = func.sum(Inventory.quantity * Journal.debit).label("quantity")
    sum_net = func.sum(Inventory.rate * Inventory.quantity * Journal.debit).label("net")
    sum_gross = func.sum(Inventory.amount * Journal.debit).label("gross")

    query = db.execute(
        select(Product, ProductGroup, sum_quantity, sum_net, sum_gross)
        .join(Inventory.batch)
        .join(Batch.sku)
        .join(StockKeepingUnit.product)
        .join(Product.product_group)
        .join(Inventory.voucher)
        .join(Voucher.journals)
        .where(
            Voucher.date >= start_date,
            Voucher.date <= finish_date,
            Voucher.voucher_type.in_([VoucherType.ISSUE, VoucherType.CLOSING_STOCK]),
            Journal.cost_centre_id == cost_centre_id,
        )
        .group_by(Product, ProductGroup, Journal.debit, ProductGroup.name)
        .order_by(ProductGroup.name, sum_net.desc())
    ).all()

    groups: Dict[uuid.UUID, schemas.RawMaterialCostItem] = {}
    counter = 0
    list_: List[schemas.RawMaterialCostItem] = []
    for product, pg, quantity, net, gross in query:
        if pg.id in groups:
            group = groups[pg.id]
            group.net += net
            group.gross += gross
        else:
            counter += 500
            group = schemas.RawMaterialCostItem(
                group=pg.name,
                net=net,
                gross=gross,
                order=counter,
                heading=True,
            )
            groups[pg.id] = group
        counter += 1
        list_.append(
            schemas.RawMaterialCostItem(
                name=product.name,
                quantity=quantity,
                net=net,
                gross=gross,
                order=counter,
                heading=False,
            )
        )

    for item in groups.values():
        list_.append(item)

    return sorted(list_, key=lambda d: d.order)
