import json
import uuid

from datetime import date, datetime
from decimal import Decimal
from typing import List, Optional, Union

from fastapi import Form
from pydantic import BaseModel, validator

from . import to_camel
from .account import AccountLink
from .batch import Batch  # noqa: F401
from .cost_centre import CostCentreLink
from .employee import EmployeeLink  # noqa: F401
from .employee_benefit import EmployeeBenefit
from .image_upload import ImageUpload
from .incentive import Incentive
from .inventory import Inventory
from .journal import Journal
from .product import ProductLink  # noqa: F401
from .user_link import UserLink


class VoucherIn(BaseModel):
    date_: date
    narration: str
    is_starred: bool
    type_: str
    files: List[ImageUpload]

    @classmethod
    def load_form(cls, data: str = Form(...)) -> BaseModel:
        json_data = json.loads(data)
        return cls.parse_obj(json_data)


class Voucher(VoucherIn):
    id_: Optional[uuid.UUID]
    is_reconciled: Optional[bool]
    reconcile_date: Optional[date]
    creation_date: Optional[datetime]
    last_edit_date: Optional[datetime]
    user: Optional[UserLink]
    posted: Optional[bool]
    poster: Optional[str]
    journals: List[Journal]
    inventories: List[Inventory]
    vendor: Optional[AccountLink]
    source: Optional[CostCentreLink]
    destination: Optional[CostCentreLink]
    incentives: List[Incentive]
    incentive: Optional[Decimal]
    employee_benefits: List[EmployeeBenefit]
    files: List[ImageUpload]

    class Config:
        anystr_strip_whitespace = True
        alias_generator = to_camel
        json_encoders = {
            date: lambda v: v.strftime("%d-%b-%Y"),
            datetime: lambda v: v.strftime("%d-%b-%Y %H:%M"),
        }

    @validator("date_", pre=True)
    def parse_date(cls, value: Union[date, str]) -> date:
        if isinstance(value, date):
            return value
        return datetime.strptime(value, "%d-%b-%Y").date()

    @validator("creation_date", pre=True)
    def parse_creation_date(cls, value: Union[None, datetime, str]) -> Optional[datetime]:
        if value is None or value == "":
            return None
        if isinstance(value, datetime):
            return value
        return datetime.strptime(value, "%d-%b-%Y %H:%M")

    @validator("last_edit_date", pre=True)
    def parse_last_edit_date(cls, value: Union[None, datetime, str]) -> Optional[datetime]:
        if value is None or value == "":
            return None
        if isinstance(value, datetime):
            return value
        return datetime.strptime(value, "%d-%b-%Y %H:%M")

    @validator("reconcile_date", pre=True)
    def parse_reconcile_date(cls, value: Union[None, date, str]) -> Optional[date]:
        if value is None or value == "":
            return None
        if isinstance(value, date):
            return value
        return datetime.strptime(value, "%d-%b-%Y").date()
